#include <string.h>
#include "OBDCan.h"

/**
 * Function Name: IsHex()
 * Description:   Checking if the character is hex digit representation
 * Returns:       True/false
 */
static bool IsHex(char ch)
{
    if (ch >= 'A' && ch <= 'F')return true;
    if (ch >= '0' && ch <= '9')return true;
    return false;
}

/**
 * Function Name: ToHex()
 * Description:   Ascii char to hex conversion
 * Returns:       Convered hex value
 */
static uint8_t ToHex(char ch)
{
    if (ch >= 'A' && ch <= 'F')return (ch - 'A' + 10);
    if (ch >= '0' && ch <= '9')return (ch - '0');
    return 0;
}

/**
 * Function Name: StringToBinary
 * Description:   Ascii/Binary conversion
 * Returns:       The buffer length
 */
static int StringToBinary(const char* str, uint8_t* buf)
{
    uint32_t len = strlen(str);

    if (len == 0 || len % 2) {
        return 0; // Invalid string
    }
    uint8_t j = 0;
    for (uint8_t i = 0; str[i]; i += 2, j++) {
        if (!IsHex(str[i]) || !IsHex(str[i+1])) {
            return 0;
        }
        buf[j]  = ToHex(str[i+1]);
        buf[j] |= ToHex(str[i]) << 4;
    }
    return j;
}

/**
 * Function Name: init()
 * Description:   Init the connection with Obd adapter
 * Returns:       The Obd adapter initialization status
 */
bool OBDCan::init(HardwareSerial *serial)
{
	char buffer[BUFFER_SIZE];
	serial_ = serial;
    serial_->setTimeout(2000); // 2sec
    //
    // reset
	//
    uint8_t sts = execCmd("ATWS\r", buffer, sizeof(buffer));
	if (sts > 0) {
		if (!strstr(buffer, "ARD1003")) {
			return false; // error
		}
		return true;
	}     
	return false;
}

/**
 * Function Name: getPidAsInt()
 * Description:   Read the PID value as integer
 * Returns:       The return value
 */
int32_t OBDCan::getPidAsInt(uint16_t pid)
{
	uint32_t val = 0;
	uint8_t buff[4] = {0};
	uint8_t num = getPidAsBytes(pid, buff, sizeof(buff));
	for (uint8_t i = 0; i < num; i++) {
		val = val << 8;
		val |= buff[i];
	}
	return num ? val : -1;
}

/**
 * Function Name: getPidAsBytes()
 * Description:   Read the PID value as sequnce of bytes
 * Returns:       The number of bytes returned or 0 if error
 */
uint8_t OBDCan::getPidAsBytes(uint16_t pid, uint8_t* buffer, uint8_t bufferSize)
{
	char cmd[REQUEST_SIZE];
	char reply[BUFFER_SIZE];
	uint8_t tmpBuff[BUFFER_SIZE / 2];

	sprintf(cmd, "%04X\r", pid);
	uint8_t sts = execCmd(cmd, reply, sizeof(reply));
	if (sts > 0) {
		uint8_t nbytes = StringToBinary(reply, tmpBuff);
		if (nbytes > 2 && (nbytes - 2) < bufferSize) { // remove extra reply bytes
			memmove(buffer, tmpBuff + 2, nbytes - 2);
			return nbytes - 2;
		}
	}
	return 0;
}

/**
 * Function Name: execCmd()
 * Description:   Request/reply OBD call
 * Returns:       Number of bytes received
 */
uint8_t OBDCan::execCmd(const char* cmd, char* outBuffer, uint8_t maxLen)
{
	send(cmd);
	return receive(outBuffer, maxLen);
}

/**
 * Function Name: send()
 * Description:   Send request to OBD chip
 * Returns:       None
 */
void OBDCan::send(const char* cmd)
{
    //com_write_log(cmd);
    serial_->write(cmd);
}

/**
 * Function Name: receive()
 * Description:   Get the OBD chip response
 * Returns:       Number of bytes received
 */
uint8_t OBDCan::receive(char* outBuffer, uint8_t maxLen)
{
	uint32_t num = serial_->readBytesUntil('>', outBuffer, maxLen);  
	if (num > 0) {
		outBuffer[num-2] = 0;
	}
	else {
		outBuffer[0] = 0;
	}
	//com_read_log(outBuffer);
	return num;
}

/**
 * Function Name: getProtocolNum()
 * Description:   Get the currect chip procotoc in use
 * Returns:       The protocol number or 0xFF if error
 */
uint8_t OBDCan::getProtocolNum()
{
	char buffer[BUFFER_SIZE];
	if (execCmd("ATDPN\r", buffer, sizeof(buffer)) > 1) {
		return buffer[0] - '0';
	}
	return 0xFF;
}